# -*- coding: utf-8 -*-

from tests import TestElasticmock, INDEX_NAME


class TestExists(TestElasticmock):

    def test_should_return_false_when_index_does_not_exists(self):
        self.assertFalse(self.es.indices.exists(INDEX_NAME))

    def test_should_return_true_when_index_exists(self):
        self.es.indices.create(INDEX_NAME)
        self.assertTrue(self.es.indices.exists(INDEX_NAME))



# -*- coding: utf-8 -*-


def for_all_methods(decorators, apply_on_public_only=True):
    def decorate(cls):
        for attr in cls.__dict__:

            if apply_on_public_only:
                if attr.startswith('_'):
                    continue

            if callable(getattr(cls, attr)):
                for decorator in decorators:
                    setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls
    return decorate



# -*- coding: utf-8 -*-

from functools import wraps

from elasticsearch.client import _normalize_hosts
from mock import patch

from elasticmock.fake_elasticsearch import FakeElasticsearch

ELASTIC_INSTANCES = {}


def _get_elasticmock(hosts=None, *args, **kwargs):
    host = _normalize_hosts(hosts)[0]
    elastic_key = '{0}:{1}'.format(
        host.get('host', 'localhost'), host.get('port', 9200)
    )

    if elastic_key in ELASTIC_INSTANCES:
        connection = ELASTIC_INSTANCES.get(elastic_key)
    else:
        connection = FakeElasticsearch()
        ELASTIC_INSTANCES[elastic_key] = connection
    return connection


def elasticmock(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        ELASTIC_INSTANCES.clear()
        with patch('elasticsearch.Elasticsearch', _get_elasticmock):
            result = f(*args, **kwargs)
        return result
    return decorated



# -*- coding: utf-8 -*-

from elasticmock import behaviour
from tests.fake_elasticsearch.behaviour import TestElasticmockBehaviour
from tests import INDEX_NAME, DOC_TYPE, BODY


class TestBehaviourServerFailure(TestElasticmockBehaviour):

    def test_should_return_internal_server_error_when_simulate_server_error_is_true(self):
        behaviour.server_failure.enable()
        data = self.es.index(index=INDEX_NAME, doc_type=DOC_TYPE, body=BODY)

        expected = {
            'status_code': 500,
            'error': 'Internal Server Error'
        }

        self.assertDictEqual(expected, data)



# -*- coding: utf-8 -*-

import unittest
from datetime import datetime

import elasticsearch

from elasticmock import elasticmock

INDEX_NAME = 'test_index'
DOC_TYPE = 'doc-Type'
DOC_ID = 'doc-id'
BODY = {
    'author': 'kimchy',
    'text': 'Elasticsearch: cool. bonsai cool.',
    'timestamp': datetime.now(),
}


class TestElasticmock(unittest.TestCase):

    @elasticmock
    def setUp(self):
        self.es = elasticsearch.Elasticsearch(hosts=[{'host': 'localhost', 'port': 9200}])



# -*- coding: utf-8 -*-

from tests import TestElasticmock, INDEX_NAME, DOC_TYPE, BODY


class TestScroll(TestElasticmock):

    def test_scrolling(self):
        for _ in range(100):
            self.es.index(index=INDEX_NAME, doc_type=DOC_TYPE, body=BODY)

        result = self.es.search(index=INDEX_NAME, params={'scroll': '1m', 'size': 30})
        self.__assert_scroll(result, 30)

        for _ in range(2):
            result = self.es.scroll(scroll_id=result.get('_scroll_id'), scroll='1m')
            self.__assert_scroll(result, 30)

        result = self.es.scroll(scroll_id=result.get('_scroll_id'), scroll='1m')
        self.__assert_scroll(result, 10)

    def __assert_scroll(self, result, expected_scroll_hits):
        hits = result.get('hits')

        self.assertNotEqual(None, result.get('_scroll_id', None))
        self.assertEqual(expected_scroll_hits, len(hits.get('hits')))
        self.assertEqual(100, hits.get('total').get('value'))



# -*- coding: utf-8 -*-

from elasticsearch.client.indices import IndicesClient
from elasticsearch.client.utils import query_params


class FakeIndicesClient(IndicesClient):

    @query_params('master_timeout', 'timeout')
    def create(self, index, body=None, params=None, headers=None):
        documents_dict = self.__get_documents_dict()
        if index not in documents_dict:
            documents_dict[index] = []

    @query_params('allow_no_indices', 'expand_wildcards', 'ignore_unavailable',
                  'local')
    def exists(self, index, params=None, headers=None):
        return index in self.__get_documents_dict()

    @query_params('allow_no_indices', 'expand_wildcards', 'force',
                  'ignore_unavailable', 'operation_threading')
    def refresh(self, index=None, params=None, headers=None):
        pass

    @query_params('master_timeout', 'timeout')
    def delete(self, index, params=None, headers=None):
        documents_dict = self.__get_documents_dict()
        if index in documents_dict:
            del documents_dict[index]

    def __get_documents_dict(self):
        return self.client._FakeElasticsearch__documents_dict



# -*- coding: utf-8 -*-

from tests import TestElasticmock


class TestHealth(TestElasticmock):

    def test_should_return_health(self):
        health_status = self.es.cluster.health()

        expected_health_status = {
            'cluster_name': 'testcluster',
            'status': 'green',
            'timed_out': False,
            'number_of_nodes': 1,
            'number_of_data_nodes': 1,
            'active_primary_shards': 1,
            'active_shards': 1,
            'relocating_shards': 0,
            'initializing_shards': 0,
            'unassigned_shards': 1,
            'delayed_unassigned_shards': 0,
            'number_of_pending_tasks': 0,
            'number_of_in_flight_fetch': 0,
            'task_max_waiting_in_queue_millis': 0,
            'active_shards_percent_as_number': 50.0
        }

        self.assertDictEqual(expected_health_status, health_status)



# -*- coding: utf-8 -*-

from __future__ import print_function

import platform

import elasticsearch

print(
    "{} {}; ElasticSearch {}".format(
        platform.python_implementation(),
        platform.python_version(),
        elasticsearch.VERSION
    )
)



# -*- coding: utf-8 -*-

from tests import TestElasticmock, INDEX_NAME


class TestRefresh(TestElasticmock):

    def test_should_refresh_index(self):
        self.es.indices.create(INDEX_NAME)
        self.es.indices.refresh(INDEX_NAME)
        self.assertTrue(self.es.indices.exists(INDEX_NAME))



# -*- coding: utf-8 -*-

from elasticsearch.client.cluster import ClusterClient
from elasticsearch.client.utils import query_params


class FakeClusterClient(ClusterClient):

    @query_params('level', 'local', 'master_timeout', 'timeout',
                  'wait_for_active_shards', 'wait_for_nodes',
                  'wait_for_relocating_shards', 'wait_for_status')
    def health(self, index=None, params=None, headers=None):
        return {
            'cluster_name': 'testcluster',
            'status': 'green',
            'timed_out': False,
            'number_of_nodes': 1,
            'number_of_data_nodes': 1,
            'active_primary_shards': 1,
            'active_shards': 1,
            'relocating_shards': 0,
            'initializing_shards': 0,
            'unassigned_shards': 1,
            'delayed_unassigned_shards': 0,
            'number_of_pending_tasks': 0,
            'number_of_in_flight_fetch': 0,
            'task_max_waiting_in_queue_millis': 0,
            'active_shards_percent_as_number': 50.0
        }



# -*- coding: utf-8 -*-

from elasticsearch.exceptions import NotFoundError

from tests import TestElasticmock, INDEX_NAME, DOC_TYPE, BODY


class TestDelete(TestElasticmock):

    def test_should_raise_exception_when_delete_nonindexed_document(self):
        with self.assertRaises(NotFoundError):
            self.es.delete(index=INDEX_NAME, doc_type=DOC_TYPE, id=1)

    def test_should_not_raise_exception_when_delete_nonindexed_document_if_ignored(self):
        target_doc = self.es.delete(index=INDEX_NAME, doc_type=DOC_TYPE, id=1, ignore=404)
        self.assertFalse(target_doc.get('found'))

    def test_should_not_raise_exception_when_delete_nonindexed_document_if_ignored_list(self):
        target_doc = self.es.delete(index=INDEX_NAME, doc_type=DOC_TYPE, id=1, ignore=(401, 404))
        self.assertFalse(target_doc.get('found'))

    def test_should_delete_indexed_document(self):
        doc_indexed = self.es.index(index=INDEX_NAME, doc_type=DOC_TYPE, body=BODY)
        search = self.es.search(index=INDEX_NAME)
        self.assertEqual(1, search.get('hits').get('total').get('value'))

        doc_id = doc_indexed.get('_id')
        doc_deleted = self.es.delete(index=INDEX_NAME, doc_type=DOC_TYPE, id=doc_id)
        search = self.es.search(index=INDEX_NAME)
        self.assertEqual(0, search.get('hits').get('total').get('value'))

        expected_doc_deleted = {
            'found': True,
            '_index': INDEX_NAME,
            '_type': DOC_TYPE,
            '_id': doc_id,
            '_version': 1,
        }

        self.assertDictEqual(expected_doc_deleted, doc_deleted)



# -*- coding: utf-8 -*-

from functools import wraps

__ENABLED = False


def enable():
    global __ENABLED
    __ENABLED = True


def disable():
    global __ENABLED
    __ENABLED = False


def server_failure(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        if __ENABLED:
            response = {
                'status_code': 500,
                'error': 'Internal Server Error'
            }
        else:
            response = f(*args, **kwargs)
        return response
    return decorated



# -*- coding: utf-8 -*-

import base64
import random
import string

DEFAULT_ELASTICSEARCH_ID_SIZE = 20
CHARSET_FOR_ELASTICSEARCH_ID = string.ascii_letters + string.digits

DEFAULT_ELASTICSEARCH_SEARCHRESULTPHASE_COUNT = 6


def get_random_id(size=DEFAULT_ELASTICSEARCH_ID_SIZE):
    return ''.join(random.choice(CHARSET_FOR_ELASTICSEARCH_ID) for _ in range(size))


def get_random_scroll_id(size=DEFAULT_ELASTICSEARCH_SEARCHRESULTPHASE_COUNT):
    return base64.b64encode(''.join(get_random_id() for _ in range(size)).encode())


def extract_ignore_as_iterable(params):
    """Extracts the value of the ignore parameter as iterable"""
    ignore = params.get('ignore', ())
    if isinstance(ignore, int):
        ignore = (ignore,)
    return ignore



# -*- coding: utf-8 -*-

import elasticsearch

from elasticmock import elasticmock
from elasticmock.fake_elasticsearch import FakeElasticsearch
from tests import TestElasticmock


class TestInstance(TestElasticmock):

    def test_should_create_fake_elasticsearch_instance(self):
        self.assertIsInstance(self.es, FakeElasticsearch)

    @elasticmock
    def test_should_return_same_elastic_instance_when_instantiate_more_than_one_instance_with_same_host(self):
        es1 = elasticsearch.Elasticsearch(hosts=[{'host': 'localhost', 'port': 9200}])
        es2 = elasticsearch.Elasticsearch(hosts=[{'host': 'localhost', 'port': 9200}])
        self.assertEqual(es1, es2)



# -*- coding: utf-8 -*-

from tests import TestElasticmock, DOC_TYPE


class TestCount(TestElasticmock):

    def test_should_return_count_for_indexed_documents_on_index(self):
        index_quantity = 0
        for i in range(0, index_quantity):
            self.es.index(index='index_{0}'.format(i), doc_type=DOC_TYPE, body={'data': 'test_{0}'.format(i)})

        count = self.es.count()
        self.assertEqual(index_quantity, count.get('count'))

    def test_should_count_in_multiple_indexes(self):
        self.es.index(index='groups', doc_type='groups', body={'budget': 1000})
        self.es.index(index='users', doc_type='users', body={'name': 'toto'})
        self.es.index(index='pcs', doc_type='pcs', body={'model': 'macbook'})

        result = self.es.count(index=['users', 'pcs'])
        self.assertEqual(2, result.get('count'))

    def test_should_count_with_empty_doc_types(self):
        self.es.index(index='index', doc_type=DOC_TYPE, body={'data': 'test'})
        count = self.es.count(doc_type=[])
        self.assertEqual(1, count.get('count'))

    def test_should_return_skipped_shards(self):
        self.es.index(index='index', doc_type=DOC_TYPE, body={'data': 'test'})
        count = self.es.count(doc_type=[])
        self.assertEqual(0, count.get('_shards').get('skipped'))

    def test_should_count_with_doc_types(self):
        self.es.index(index='index', doc_type=DOC_TYPE, body={'data': 'test1'})
        self.es.index(index='index', doc_type='different-doc-type', body={'data': 'test2'})
        count = self.es.count(doc_type=DOC_TYPE)
        self.assertEqual(1, count.get('count'))



# -*- coding: utf-8 -*-

import json

from tests import TestElasticmock, INDEX_NAME, DOC_TYPE, BODY, DOC_ID


class TestBulk(TestElasticmock):

    def test_should_bulk_index_documents_index_creates(self):
        action = {'index': {'_index': INDEX_NAME, '_type': DOC_TYPE}}
        action_json = json.dumps(action)
        body_json = json.dumps(BODY, default=str)
        num_of_documents = 10

        lines = []
        for count in range(0, num_of_documents):
            lines.append(action_json)
            lines.append(body_json)
        body = '\n'.join(lines)

        data = self.es.bulk(body=body)
        items = data.get('items')

        self.assertFalse(data.get('errors'))
        self.assertEqual(num_of_documents, len(items))

        for item in items:
            index = item.get('index')
            self.assertEqual(DOC_TYPE, index.get('_type'))
            self.assertEqual(INDEX_NAME, index.get('_index'))
            self.assertEqual('created', index.get('result'))
            self.assertEqual(201, index.get('status'))

    def test_should_bulk_index_documents_create_creates(self):
        create_action = {'create': {'_index': INDEX_NAME, '_type': DOC_TYPE}}
        create_with_id = {'create': {'_index': INDEX_NAME, '_type': DOC_TYPE, '_id': DOC_ID}}
        actions = [
            json.dumps(create_action),
            json.dumps(BODY, default=str),
            json.dumps(create_action),
            json.dumps(BODY, default=str),
            json.dumps(create_with_id),
            json.dumps(BODY, default=str),
            # Will fail on created document with the same ID
            json.dumps(create_with_id),
            json.dumps(BODY, default=str),
        ]
        body = '\n'.join(actions)

        data = self.es.bulk(body=body)

        items = data.get('items')

        self.assertTrue(data.get('errors'))
        self.assertEqual(4, len(items))

        last_item = items.pop()
        self.assertEqual(last_item['create']['error'], 'version_conflict_engine_exception')
        self.assertEqual(last_item['create']['status'], 409)
        for item in items:
            index = item.get('create')
            self.assertEqual(DOC_TYPE, index.get('_type'))
            self.assertEqual(INDEX_NAME, index.get('_index'))
            self.assertEqual('created', index.get('result'))
            self.assertEqual(201, index.get('status'))

    def test_should_bulk_index_documents_index_updates(self):
        action = {'index': {'_index': INDEX_NAME, '_id': DOC_ID, '_type': DOC_TYPE}}
        action_json = json.dumps(action)
        body_json = json.dumps(BODY, default=str)
        num_of_documents = 10

        lines = []
        for count in range(0, num_of_documents):
            lines.append(action_json)
            lines.append(body_json)
        body = '\n'.join(lines)

        data = self.es.bulk(body=body)
        items = data.get('items')

        self.assertFalse(data.get('errors'))
        self.assertEqual(num_of_documents, len(items))

        first_item = items.pop(0)
        self.assertEqual(first_item["index"]["status"], 201)
        self.assertEqual(first_item["index"]["result"], "created")

        for item in items:
            index = item.get('index')
            self.assertEqual(DOC_TYPE, index.get('_type'))
            self.assertEqual(INDEX_NAME, index.get('_index'))
            self.assertEqual('updated', index.get('result'))
            self.assertEqual(200, index.get('status'))

    def test_should_bulk_index_documents_update_updates(self):
        action = {'update': {'_index': INDEX_NAME, '_id': DOC_ID, '_type': DOC_TYPE}}
        action_json = json.dumps(action)
        create_action_json = json.dumps(
            {'create': {'_index': INDEX_NAME, '_id': DOC_ID, '_type': DOC_TYPE}}
        )
        body_json = json.dumps({'doc': BODY}, default=str)
        num_of_documents = 4

        lines = [create_action_json, json.dumps(BODY, default=str)]
        for count in range(0, num_of_documents):
            lines.append(action_json)
            lines.append(body_json)
        body = '\n'.join(lines)

        data = self.es.bulk(body=body)
        items = data.get('items')

        self.assertFalse(data.get('errors'))
        self.assertEqual(num_of_documents + 1, len(items))

        first_item = items.pop(0)
        self.assertEqual(first_item["create"]["status"], 201)
        self.assertEqual(first_item["create"]["result"], "created")

        for item in items:
            index = item.get('update')
            self.assertEqual(DOC_TYPE, index.get('_type'))
            self.assertEqual(INDEX_NAME, index.get('_index'))
            self.assertEqual('updated', index.get('result'))
            self.assertEqual(200, index.get('status'))

    def test_should_bulk_index_documents_delete_deletes(self):
        delete_action = {'delete': {'_index': INDEX_NAME, '_id': DOC_ID, '_type': DOC_TYPE}}
        delete_action_json = json.dumps(delete_action)
        create_action_json = json.dumps(
            {'create': {'_index': INDEX_NAME, '_id': DOC_ID, '_type': DOC_TYPE}}
        )

        lines = [
            create_action_json,
            json.dumps(BODY, default=str),
            delete_action_json,
        ]
        body = '\n'.join(lines)

        data = self.es.bulk(body=body)
        items = data.get('items')

        self.assertFalse(data.get('errors'))
        self.assertEqual(2, len(items))

        first_item = items.pop(0)
        self.assertEqual(first_item["create"]["status"], 201)
        self.assertEqual(first_item["create"]["result"], "created")
        self.assertEqual(first_item["create"]['_type'], DOC_TYPE)
        self.assertEqual(first_item["create"]['_id'], DOC_ID)

        second_item = items.pop(0)
        self.assertEqual(second_item["delete"]["status"], 200)
        self.assertEqual(second_item["delete"]["result"], "deleted")
        self.assertEqual(second_item["delete"]['_type'], DOC_TYPE)
        self.assertEqual(second_item["delete"]['_id'], DOC_ID)

    def test_should_bulk_index_documents_mixed_actions(self):
        doc_body = json.dumps(BODY, default=str)

        doc_id_1 = 1
        doc_id_2 = 2
        actions = [
            json.dumps({'create': {'_index': INDEX_NAME, '_type': DOC_TYPE, '_id': doc_id_1}}),
            doc_body,  # 201
            json.dumps({'create': {'_index': INDEX_NAME, '_type': DOC_TYPE, '_id': doc_id_1}}),
            doc_body,  # 409
            json.dumps({'index': {'_index': INDEX_NAME, '_type': DOC_TYPE, '_id': doc_id_2}}),
            doc_body,  # 201
            json.dumps({'index': {'_index': INDEX_NAME, '_type': DOC_TYPE, '_id': doc_id_2}}),
            doc_body,  # 200
            json.dumps({'update': {'_index': INDEX_NAME, '_type': DOC_TYPE, '_id': doc_id_1}}),
            doc_body,  # 200
            json.dumps({'delete': {'_index': INDEX_NAME, '_type': DOC_TYPE, '_id': doc_id_1}}),
            # 200
            json.dumps({'update': {'_index': INDEX_NAME, '_type': DOC_TYPE, '_id': doc_id_1}}),
            doc_body,  # 404
            json.dumps({'delete': {'_index': INDEX_NAME, '_type': DOC_TYPE, '_id': doc_id_1}}),
            # 404
        ]
        body = '\n'.join(actions)

        data = self.es.bulk(body=body)

        expected = [
            {'create': {'_type': DOC_TYPE, '_id': 1, '_index': INDEX_NAME,
                        '_version': 1, 'status': 201, 'result': 'created'}},
            {'create': {'_type': DOC_TYPE, '_id': 1, '_index': INDEX_NAME,
                        '_version': 1, 'status': 409,
                        'error': 'version_conflict_engine_exception'}},
            {'index': {'_type': DOC_TYPE, '_id': 2, '_index': INDEX_NAME,
                       '_version': 1, 'status': 201, 'result': 'created'}},
            {'index': {'_type': DOC_TYPE, '_id': 2, '_index': INDEX_NAME,
                       '_version': 1, 'status': 200, 'result': 'updated'}},
            {'update': {'_type': DOC_TYPE, '_id': 1, '_index': INDEX_NAME,
                        '_version': 1, 'status': 200, 'result': 'updated'}},
            {'delete': {'_type': DOC_TYPE, '_id': 1, '_index': INDEX_NAME,
                        '_version': 1, 'result': 'deleted', 'status': 200}},
            {'update': {'_type': DOC_TYPE, '_id': 1, '_index': INDEX_NAME,
                        '_version': 1, 'status': 404, 'error': 'document_missing_exception'}},
            {'delete': {'_type': DOC_TYPE, '_id': 1, '_index': INDEX_NAME,
                        '_version': 1, 'error': 'not_found', 'status': 404}},
            ]

        actual = data.get('items')

        self.assertTrue(data.get('errors'))
        self.assertEqual(actual, expected)



# -*- coding: utf-8 -*-

from tests import TestElasticmock, INDEX_NAME, DOC_TYPE, BODY

UPDATED_BODY = {
    'author': 'vrcmarcos',
    'text': 'Updated Text'
}


class TestIndex(TestElasticmock):

    def test_should_index_document(self):
        data = self.es.index(index=INDEX_NAME, doc_type=DOC_TYPE, body=BODY)

        self.assertEqual(DOC_TYPE, data.get('_type'))
        self.assertTrue(data.get('created'))
        self.assertEqual(1, data.get('_version'))
        self.assertEqual(INDEX_NAME, data.get('_index'))

    def test_should_index_document_without_doc_type(self):
        data = self.es.index(index=INDEX_NAME, body=BODY)

        self.assertEqual('_doc', data.get('_type'))
        self.assertTrue(data.get('created'))
        self.assertEqual(1, data.get('_version'))
        self.assertEqual(INDEX_NAME, data.get('_index'))

    def test_doc_type_can_be_list(self):
        doc_types = ['1_idx', '2_idx', '3_idx']
        count_per_doc_type = 3

        for doc_type in doc_types:
            for _ in range(count_per_doc_type):
                self.es.index(index=INDEX_NAME, doc_type=doc_type, body={})

        result = self.es.search(doc_type=[doc_types[0]])
        self.assertEqual(count_per_doc_type, result.get('hits').get('total').get('value'))

        result = self.es.search(doc_type=doc_types[:2])
        self.assertEqual(count_per_doc_type * 2, result.get('hits').get('total').get('value'))

    def test_update_existing_doc(self):
        data = self.es.index(index=INDEX_NAME, doc_type=DOC_TYPE, body=BODY)
        document_id = data.get('_id')
        self.es.index(index=INDEX_NAME, id=document_id, doc_type=DOC_TYPE, body=UPDATED_BODY)
        target_doc = self.es.get(index=INDEX_NAME, id=document_id)

        expected = {
            '_type': DOC_TYPE,
            '_source': UPDATED_BODY,
            '_index': INDEX_NAME,
            '_version': 2,
            'found': True,
            '_id': document_id
        }

        self.assertDictEqual(expected, target_doc)



# -*- coding: utf-8 -*-

from tests import TestElasticmock, INDEX_NAME, DOC_TYPE, BODY


class TestExists(TestElasticmock):

    def test_should_return_exists_false_if_nonindexed_id_is_used(self):
        self.assertFalse(self.es.exists(index=INDEX_NAME, doc_type=DOC_TYPE, id=1))

    def test_should_return_exists_true_if_indexed_id_is_used(self):
        data = self.es.index(index=INDEX_NAME, doc_type=DOC_TYPE, body=BODY)
        document_id = data.get('_id')
        self.assertTrue(self.es.exists(index=INDEX_NAME, doc_type=DOC_TYPE, id=document_id))



# -*- coding: utf-8 -*-

from elasticsearch.exceptions import NotFoundError

from tests import TestElasticmock, INDEX_NAME, DOC_TYPE, BODY


class TestSuggest(TestElasticmock):

    def test_should_raise_notfounderror_when_nonindexed_id_is_used_for_suggest(self):
        with self.assertRaises(NotFoundError):
            self.es.suggest(body={}, index=INDEX_NAME)

    def test_should_return_suggestions(self):
        self.es.index(index=INDEX_NAME, doc_type=DOC_TYPE, body=BODY)
        suggestion_body = {
            'suggestion-string': {
                'text': 'test_text',
                'term': {
                    'field': 'string'
                }
            },
            'suggestion-id': {
                'text': 1234567,
                'term': {
                    'field': 'id'
                }
            }
        }
        suggestion = self.es.suggest(body=suggestion_body, index=INDEX_NAME)
        self.assertIsNotNone(suggestion)
        self.assertDictEqual({
            'suggestion-string': [
                {
                    'text': 'test_text',
                    'length': 1,
                    'options': [
                        {
                            'text': 'test_text_suggestion',
                            'freq': 1,
                            'score': 1.0
                        }
                    ],
                    'offset': 0
                }
            ],
            'suggestion-id': [
                {
                    'text': 1234567,
                    'length': 1,
                    'options': [
                        {
                            'text': 1234568,
                            'freq': 1,
                            'score': 1.0
                        }
                    ],
                    'offset': 0
                }
            ],
        }, suggestion)



# -*- coding: utf-8 -*-

from tests import TestElasticmock


class TestInfo(TestElasticmock):

    def test_should_return_status_200_for_info(self):
        info = self.es.info()
        self.assertEqual(info.get('status'), 200)



# -*- coding: utf-8 -*-
import datetime

from elasticsearch.exceptions import NotFoundError
from parameterized import parameterized

from tests import TestElasticmock, INDEX_NAME, DOC_TYPE


class TestSearch(TestElasticmock):

    def test_should_raise_notfounderror_when_search_for_unexistent_index(self):
        with self.assertRaises(NotFoundError):
            self.es.search(index=INDEX_NAME)

    def test_should_return_hits_hits_even_when_no_result(self):
        search = self.es.search()
        self.assertEqual(0, search.get('hits').get('total').get('value'))
        self.assertListEqual([], search.get('hits').get('hits'))

    def test_should_return_skipped_shards(self):
        search = self.es.search()
        self.assertEqual(0, search.get('_shards').get('skipped'))

    def test_should_return_all_documents(self):
        index_quantity = 10
        for i in range(0, index_quantity):
            self.es.index(index='index_{0}'.format(i), doc_type=DOC_TYPE, body={'data': 'test_{0}'.format(i)})

        search = self.es.search()
        self.assertEqual(index_quantity, search.get('hits').get('total').get('value'))

    def test_should_return_all_documents_match_all(self):
        index_quantity = 10
        for i in range(0, index_quantity):
            self.es.index(index='index_{0}'.format(i), doc_type=DOC_TYPE, body={'data': 'test_{0}'.format(i)})

        search = self.es.search(body={'query': {'match_all': {}}})
        self.assertEqual(index_quantity, search.get('hits').get('total').get('value'))

    def test_should_return_only_indexed_documents_on_index(self):
        index_quantity = 2
        for i in range(0, index_quantity):
            self.es.index(index=INDEX_NAME, doc_type=DOC_TYPE, body={'data': 'test_{0}'.format(i)})

        search = self.es.search(index=INDEX_NAME)
        self.assertEqual(index_quantity, search.get('hits').get('total').get('value'))

    def test_should_return_only_indexed_documents_on_index_with_doc_type(self):
        index_quantity = 2
        for i in range(0, index_quantity):
            self.es.index(index=INDEX_NAME, doc_type=DOC_TYPE, body={'data': 'test_{0}'.format(i)})
        self.es.index(index=INDEX_NAME, doc_type='another-Doctype', body={'data': 'test'})

        search = self.es.search(index=INDEX_NAME, doc_type=DOC_TYPE)
        self.assertEqual(index_quantity, search.get('hits').get('total').get('value'))

    def test_should_search_in_multiple_indexes(self):
        self.es.index(index='groups', doc_type='groups', body={'budget': 1000})
        self.es.index(index='users', doc_type='users', body={'name': 'toto'})
        self.es.index(index='pcs', doc_type='pcs', body={'model': 'macbook'})

        result = self.es.search(index=['users', 'pcs'])
        self.assertEqual(2, result.get('hits').get('total').get('value'))

    def test_usage_of_aggregations(self):
        self.es.index(index='index', doc_type='document', body={'genre': 'rock'})

        body = {"aggs": {"genres": {"terms": {"field": "genre"}}}}
        result = self.es.search(index='index', body=body)

        self.assertTrue('aggregations' in result)

    def test_search_with_scroll_param(self):
        for _ in range(100):
            self.es.index(index='groups', doc_type='groups', body={'budget': 1000})

        result = self.es.search(index='groups', params={'scroll': '1m', 'size': 30})
        self.assertNotEqual(None, result.get('_scroll_id', None))
        self.assertEqual(30, len(result.get('hits').get('hits')))
        self.assertEqual(100, result.get('hits').get('total').get('value'))

    def test_search_with_match_query(self):
        for i in range(0, 10):
            self.es.index(index='index_for_search', doc_type=DOC_TYPE, body={'data': 'test_{0}'.format(i)})

        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE,
                                  body={'query': {'match': {'data': 'TEST'}}})
        self.assertEqual(response['hits']['total']['value'], 10)
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 10)

        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE, body={'query': {'match': {'data': '3'}}})
        self.assertEqual(response['hits']['total']['value'], 1)
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 1)
        self.assertEqual(hits[0]['_source'], {'data': 'test_3'})

    def test_search_with_match_query_in_int_list(self):
        for i in range(0, 10):
            self.es.index(index='index_for_search', doc_type=DOC_TYPE, body={'data': [i, 11, 13]})
        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE, body={'query': {'match': {'data': 1}}})
        self.assertEqual(response['hits']['total']['value'], 1)
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 1)
        self.assertEqual(hits[0]['_source'], {'data': [1, 11, 13]})

    def test_search_with_match_query_in_string_list(self):
        for i in range(0, 10):
            self.es.index(index='index_for_search', doc_type=DOC_TYPE, body={'data': [str(i), 'two', 'three']})

        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE, body={'query': {'match': {'data': '1'}}})
        self.assertEqual(response['hits']['total']['value'], 1)
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 1)
        self.assertEqual(hits[0]['_source'], {'data': ['1', 'two', 'three']})

    def test_search_with_term_query(self):
        for i in range(0, 10):
            self.es.index(index='index_for_search', doc_type=DOC_TYPE, body={'data': 'test_{0}'.format(i)})

        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE,
                                  body={'query': {'term': {'data': 'TEST'}}})
        self.assertEqual(response['hits']['total']['value'], 0)
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 0)

        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE, body={'query': {'term': {'data': '3'}}})
        self.assertEqual(response['hits']['total']['value'], 1)
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 1)
        self.assertEqual(hits[0]['_source'], {'data': 'test_3'})

    def test_search_with_bool_query(self):
        for i in range(0, 10):
            self.es.index(index='index_for_search', doc_type=DOC_TYPE, body={'id': i})

        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE,
                                  body={'query': {'bool': {'filter': [{'term': {'id': 1}}]}}})
        self.assertEqual(response['hits']['total']['value'], 1)
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 1)

    def test_search_with_must_not_query(self):
        for i in range(0, 10):
            self.es.index(index='index_for_search', doc_type=DOC_TYPE, body={'id': i})
        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE,
                                  body={'query': {'bool': {
                                      'filter': [{'terms': {'id': [1, 2]}}],
                                      'must_not': [{'term': {'id': 1}}],
                                      }}})
        self.assertEqual(response['hits']['total']['value'], 1)
        doc = response['hits']['hits'][0]['_source']
        self.assertEqual(2, doc['id'])

    def test_search_with_terms_query(self):
        for i in range(0, 10):
            self.es.index(index='index_for_search', doc_type=DOC_TYPE, body={'id': i})

        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE,
                                  body={'query': {'terms': {'id': [1, 2, 3]}}})
        self.assertEqual(response['hits']['total']['value'], 3)
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 3)

    def test_query_on_nested_data(self):
        for i, y in enumerate(['yes', 'no']):
            self.es.index('index_for_search', doc_type=DOC_TYPE,
                          body={'id': i, 'data': {'x': i, 'y': y}})

        for term, value, i in [('data.x', 1, 1), ('data.y', 'yes', 0)]:
            response = self.es.search(index='index_for_search', doc_type=DOC_TYPE,
                                      body={'query': {'term': {term: value}}})
            self.assertEqual(1, response['hits']['total']['value'])
            doc = response['hits']['hits'][0]['_source']
            self.assertEqual(i, doc['id'])

    def test_search_with_bool_query_and_multi_match(self):
        for i in range(0, 10):
            self.es.index(index='index_for_search', doc_type=DOC_TYPE, body={
                'data': 'test_{0}'.format(i) if i % 2 == 0 else None,
                'data2': 'test_{0}'.format(i) if (i+1) % 2 == 0 else None
                })

        search_body = {
            "query": {
                "bool": {
                    "must": {
                        "multi_match": {
                            "query": "test",
                            "fields": ["data", "data2"]
                        }
                    }
                }
            }
        }
        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE,
                                  body=search_body)
        self.assertEqual(response['hits']['total']['value'], 10)
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 10)

    def test_search_bool_should_match_query(self):
        for i in range(0, 10):
            self.es.index(index='index_for_search', doc_type=DOC_TYPE, body={'data': 'test_{0}'.format(i)})

        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE,
                                  body={
                                      'query': {
                                          'bool': {
                                              'should': [
                                                  {'match': {'data': 'test_0'}},
                                                  {'match': {'data': 'test_1'}},
                                                  {'match': {'data': 'test_2'}},
                                              ]
                                          }
                                      }
                                  })
        self.assertEqual(response['hits']['total']['value'], 3)
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 3)
        self.assertEqual(hits[0]['_source'], {'data': 'test_0'})

    def test_msearch(self):
        for i in range(0, 10):
            self.es.index(index='index_for_search1', doc_type=DOC_TYPE, body={
                'data': 'test_{0}'.format(i) if i % 2 == 0 else None,
                'data2': 'test_{0}'.format(i) if (i+1) % 2 == 0 else None
                })
        for i in range(0, 10):
            self.es.index(index='index_for_search2', doc_type=DOC_TYPE, body={
                'data': 'test_{0}'.format(i) if i % 2 == 0 else None,
                'data2': 'test_{0}'.format(i) if (i+1) % 2 == 0 else None
                })

        search_body = {
            "query": {
                "bool": {
                    "must": {
                        "multi_match": {
                            "query": "test",
                            "fields": ["data", "data2"]
                        }
                    }
                }
            }
        }
        body = []
        body.append({'index': 'index_for_search1'})
        body.append(search_body)
        body.append({'index': 'index_for_search2'})
        body.append(search_body)

        result = self.es.msearch(index='index_for_search', body=body)
        response1, response2 = result['responses']
        self.assertEqual(response1['hits']['total']['value'], 10)
        hits1 = response1['hits']['hits']
        self.assertEqual(len(hits1), 10)
        self.assertEqual(response2['hits']['total']['value'], 10)
        hits2 = response2['hits']['hits']
        self.assertEqual(len(hits2), 10)

    def test_search_with_sort(self):
        for i in range(0, 10):
            self.es.index(index='index_for_search', doc_type=DOC_TYPE, body={'id': i})
        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE,
                                  body={
                                      "query": {"match_all": {}},
                                      "sort": ["id"]
                                      })
        self.assertEqual(response['hits']['total']['value'], 10)
        doc = response['hits']['hits'][0]['_source']
        self.assertEqual(0, doc['id'])

        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE,
                                  body={
                                      "query": {"match_all": {}},
                                      "sort": [{"id": "desc"}]
                                      })
        self.assertEqual(response['hits']['total']['value'], 10)
        doc = response['hits']['hits'][0]['_source']
        self.assertEqual(9, doc['id'])

        response = self.es.search(index='index_for_search', doc_type=DOC_TYPE,
                                  body={
                                      "query": {"match_all": {}},
                                      "sort": [{"id": {"order": "desc"}}]
                                      })
        self.assertEqual(response['hits']['total']['value'], 10)
        doc = response['hits']['hits'][0]['_source']
        self.assertEqual(9, doc['id'])

    @parameterized.expand(
        [
            (
                    'timestamp gt',
                    {'timestamp': {'gt': datetime.datetime(2009, 1, 1, 10, 20, 0).isoformat()}},
                    range(5, 12),
            ),
            (
                    'timestamp gte',
                    {'timestamp': {'gte': datetime.datetime(2009, 1, 1, 10, 20, 0).isoformat()}},
                    range(4, 12),
            ),
            (
                    'timestamp lt',
                    {'timestamp': {'lt': datetime.datetime(2009, 1, 1, 10, 35, 0).isoformat()}},
                    range(7),
            ),
            (
                    'timestamp lte',
                    {'timestamp': {'lte': datetime.datetime(2009, 1, 1, 10, 35, 0).isoformat()}},
                    range(8),
            ),
            (
                    'timestamp combination',
                    {
                        'timestamp': {
                            'gt': datetime.datetime(2009, 1, 1, 10, 15, 0).isoformat(),
                            'lte': datetime.datetime(2009, 1, 1, 10, 35, 0).isoformat(),
                        }
                    },
                    range(4, 8),
            ),
            (
                    'data_int gt',
                    {'data_int': {'gt': 40}},
                    range(5, 12),
            ),
            (
                    'data_int gte',
                    {'data_int': {'gte': 40}},
                    range(4, 12),
            ),
            (
                    'data_int lt',
                    {'data_int': {'lt': 70}},
                    range(7),
            ),
            (
                    'data_int lte',
                    {'data_int': {'lte': 70}},
                    range(8),
            ),
            (
                    'data_int combination',
                    {'data_int': {'gt': 30, 'lte': 70}},
                    range(4, 8),
            ),
        ]
    )
    def test_search_with_range_query(self, _, query_range, expected_ids):
        for i in range(0, 12):
            body = {
                'id': i,
                'timestamp': datetime.datetime(2009, 1, 1, 10, 5 * i, 0),
                'data_int': 10 * i,
            }
            self.es.index(index='index_for_search', doc_type=DOC_TYPE, body=body)

        response = self.es.search(
            index='index_for_search',
            doc_type=DOC_TYPE,
            body={'query': {'range': query_range}},
        )

        self.assertEqual(len(expected_ids), response['hits']['total']['value'])
        hits = response['hits']['hits']
        self.assertEqual(set(expected_ids), set(hit['_source']['id'] for hit in hits))

    def test_bucket_aggregation(self):
        data = [
            {"data_x": 1, "data_y": "a"},
            {"data_x": 1, "data_y": "a"},
            {"data_x": 2, "data_y": "a"},
            {"data_x": 2, "data_y": "b"},
            {"data_x": 3, "data_y": "b"},
        ]
        for body in data:
            self.es.index(index='index_for_search', doc_type=DOC_TYPE, body=body)

        response = self.es.search(
            index="index_for_search",
            doc_type=DOC_TYPE,
            body={
                "query": {"match_all": {}},
                "aggs": {
                    "stats": {
                        "composite": {
                            "sources": [{"data_x": {"terms": {"field": "data_x"}}}],
                            "size": 10000,
                        },
                        "aggs": {
                            "distinct_data_y": {"cardinality": {"field": "data_y"}}
                        },
                    }
                },
            },
        )

        expected = [
            {"key": {"data_x": 1}, "doc_count": 2},
            {"key": {"data_x": 2}, "doc_count": 2},
            {"key": {"data_x": 3}, "doc_count": 1},
        ]
        actual = response["aggregations"]["stats"]["buckets"]

        for x, y in zip(expected, actual):
            self.assertDictEqual(x["key"], y["key"])
            self.assertEqual(x["doc_count"], y["doc_count"])



# -*- coding: utf-8 -*-
import datetime
import json
import sys
from collections import defaultdict

import dateutil.parser
from elasticsearch import Elasticsearch
from elasticsearch.client.utils import query_params
from elasticsearch.client import _normalize_hosts
from elasticsearch.transport import Transport
from elasticsearch.exceptions import NotFoundError, RequestError

from elasticmock.behaviour.server_failure import server_failure
from elasticmock.fake_cluster import FakeClusterClient
from elasticmock.fake_indices import FakeIndicesClient
from elasticmock.utilities import (extract_ignore_as_iterable, get_random_id,
    get_random_scroll_id)
from elasticmock.utilities.decorator import for_all_methods

PY3 = sys.version_info[0] == 3
if PY3:
    unicode = str


class QueryType:
    BOOL = 'BOOL'
    FILTER = 'FILTER'
    MATCH = 'MATCH'
    MATCH_ALL = 'MATCH_ALL'
    TERM = 'TERM'
    TERMS = 'TERMS'
    MUST = 'MUST'
    RANGE = 'RANGE'
    SHOULD = 'SHOULD'
    MINIMUM_SHOULD_MATCH = 'MINIMUM_SHOULD_MATCH'
    MULTI_MATCH = 'MULTI_MATCH'
    MUST_NOT = 'MUST_NOT'

    @staticmethod
    def get_query_type(type_str):
        if type_str == 'bool':
            return QueryType.BOOL
        elif type_str == 'filter':
            return QueryType.FILTER
        elif type_str == 'match':
            return QueryType.MATCH
        elif type_str == 'match_all':
            return QueryType.MATCH_ALL
        elif type_str == 'term':
            return QueryType.TERM
        elif type_str == 'terms':
            return QueryType.TERMS
        elif type_str == 'must':
            return QueryType.MUST
        elif type_str == 'range':
            return QueryType.RANGE
        elif type_str == 'should':
            return QueryType.SHOULD
        elif type_str == 'minimum_should_match':
            return QueryType.MINIMUM_SHOULD_MATCH
        elif type_str == 'multi_match':
            return QueryType.MULTI_MATCH
        elif type_str == 'must_not':
            return QueryType.MUST_NOT
        else:
            raise NotImplementedError(f'type {type_str} is not implemented for QueryType')


class MetricType:
    CARDINALITY = "CARDINALITY"

    @staticmethod
    def get_metric_type(type_str):
        if type_str == "cardinality":
            return MetricType.CARDINALITY
        else:
            raise NotImplementedError(f'type {type_str} is not implemented for MetricType')


class FakeQueryCondition:
    type = None
    condition = None

    def __init__(self, type, condition):
        self.type = type
        self.condition = condition

    def evaluate(self, document):
        return self._evaluate_for_query_type(document)

    def _evaluate_for_query_type(self, document):
        if self.type == QueryType.MATCH:
            return self._evaluate_for_match_query_type(document)
        elif self.type == QueryType.MATCH_ALL:
            return True
        elif self.type == QueryType.TERM:
            return self._evaluate_for_term_query_type(document)
        elif self.type == QueryType.TERMS:
            return self._evaluate_for_terms_query_type(document)
        elif self.type == QueryType.RANGE:
            return self._evaluate_for_range_query_type(document)
        elif self.type == QueryType.BOOL:
            return self._evaluate_for_compound_query_type(document)
        elif self.type == QueryType.FILTER:
            return self._evaluate_for_compound_query_type(document)
        elif self.type == QueryType.MUST:
            return self._evaluate_for_compound_query_type(document)
        elif self.type == QueryType.SHOULD:
            return self._evaluate_for_should_query_type(document)
        elif self.type == QueryType.MULTI_MATCH:
            return self._evaluate_for_multi_match_query_type(document)
        elif self.type == QueryType.MUST_NOT:
            return self._evaluate_for_must_not_query_type(document)
        else:
            raise NotImplementedError('Fake query evaluation not implemented for query type: %s' % self.type)

    def _evaluate_for_match_query_type(self, document):
        return self._evaluate_for_field(document, True)

    def _evaluate_for_term_query_type(self, document):
        return self._evaluate_for_field(document, False)

    def _evaluate_for_terms_query_type(self, document):
        for field in self.condition:
            for term in self.condition[field]:
                if FakeQueryCondition(QueryType.TERM, {field: term}).evaluate(document):
                    return True
        return False

    def _evaluate_for_field(self, document, ignore_case):
        doc_source = document['_source']
        return_val = False
        for field, value in self.condition.items():
            return_val = self._compare_value_for_field(
                doc_source,
                field,
                value,
                ignore_case
            )
            if return_val:
                break
        return return_val

    def _evaluate_for_fields(self, document):
        doc_source = document['_source']
        return_val = False
        value = self.condition.get('query')
        if not value:
            return return_val
        fields = self.condition.get('fields', [])
        for field in fields:
            return_val = self._compare_value_for_field(
                doc_source,
                field,
                value,
                True
            )
            if return_val:
                break

        return return_val

    def _evaluate_for_range_query_type(self, document):
        for field, comparisons in self.condition.items():
            doc_val = document['_source']
            for k in field.split("."):
                if hasattr(doc_val, k):
                    doc_val = getattr(doc_val, k)
                elif k in doc_val:
                    doc_val = doc_val[k]
                else:
                    return False

            if isinstance(doc_val, list):
                return False

            for sign, value in comparisons.items():
                if isinstance(doc_val, datetime.datetime):
                    value = dateutil.parser.isoparse(value)
                if sign == 'gte':
                    if doc_val < value:
                        return False
                elif sign == 'gt':
                    if doc_val <= value:
                        return False
                elif sign == 'lte':
                    if doc_val > value:
                        return False
                elif sign == 'lt':
                    if doc_val >= value:
                        return False
                else:
                    raise ValueError(f"Invalid comparison type {sign}")
            return True

    def _evaluate_for_compound_query_type(self, document):
        return_val = False
        if isinstance(self.condition, dict):
            for query_type, sub_query in self.condition.items():
                return_val = FakeQueryCondition(
                    QueryType.get_query_type(query_type),
                    sub_query
                ).evaluate(document)
                if not return_val:
                    return False
        elif isinstance(self.condition, list):
            for sub_condition in self.condition:
                for sub_condition_key in sub_condition:
                    return_val = FakeQueryCondition(
                        QueryType.get_query_type(sub_condition_key),
                        sub_condition[sub_condition_key]
                    ).evaluate(document)
                    if not return_val:
                        return False

        return return_val

    def _evaluate_for_must_not_query_type(self, document):
        if isinstance(self.condition, dict):
            for query_type, sub_query in self.condition.items():
                return_val = FakeQueryCondition(
                    QueryType.get_query_type(query_type),
                    sub_query
                ).evaluate(document)
                if return_val:
                    return False
        elif isinstance(self.condition, list):
            for sub_condition in self.condition:
                for sub_condition_key in sub_condition:
                    return_val = FakeQueryCondition(
                        QueryType.get_query_type(sub_condition_key),
                        sub_condition[sub_condition_key]
                    ).evaluate(document)
                    if return_val:
                        return False
        return True

    def _evaluate_for_should_query_type(self, document):
        return_val = False
        for sub_condition in self.condition:
            for sub_condition_key in sub_condition:
                return_val = FakeQueryCondition(
                    QueryType.get_query_type(sub_condition_key),
                    sub_condition[sub_condition_key]
                ).evaluate(document)
                if return_val:
                    return True
        return return_val

    def _evaluate_for_multi_match_query_type(self, document):
        return self._evaluate_for_fields(document)

    def _compare_value_for_field(self, doc_source, field, value, ignore_case):
        if ignore_case and isinstance(value, str):
            value = value.lower()

        doc_val = doc_source
        # Remove boosting
        field, *_ = field.split("*")
        for k in field.split("."):
            if hasattr(doc_val, k):
                doc_val = getattr(doc_val, k)
                break
            elif k in doc_val:
                doc_val = doc_val[k]
                break
            else:
                return False

        if not isinstance(doc_val, list):
            doc_val = [doc_val]

        for val in doc_val:
            if not isinstance(val, (int, float, complex)) or val is None:
                val = str(val)
                if ignore_case:
                    val = val.lower()

            if value == val:
                return True
            if isinstance(val, str) and str(value) in val:
                return True

        return False


@for_all_methods([server_failure])
class FakeElasticsearch(Elasticsearch):
    __documents_dict = None

    def __init__(self, hosts=None, transport_class=None, **kwargs):
        self.__documents_dict = {}
        self.__scrolls = {}
        self.transport = Transport(_normalize_hosts(hosts), **kwargs)

    @property
    def indices(self):
        return FakeIndicesClient(self)

    @property
    def cluster(self):
        return FakeClusterClient(self)

    @query_params()
    def ping(self, params=None, headers=None):
        return True

    @query_params()
    def info(self, params=None, headers=None):
        return {
            'status': 200,
            'cluster_name': 'elasticmock',
            'version':
                {
                    'lucene_version': '4.10.4',
                    'build_hash': '00f95f4ffca6de89d68b7ccaf80d148f1f70e4d4',
                    'number': '1.7.5',
                    'build_timestamp': '2016-02-02T09:55:30Z',
                    'build_snapshot': False
                },
            'name': 'Nightwatch',
            'tagline': 'You Know, for Search'
        }

    @query_params('consistency',
                  'op_type',
                  'parent',
                  'refresh',
                  'replication',
                  'routing',
                  'timeout',
                  'timestamp',
                  'ttl',
                  'version',
                  'version_type')
    def index(self, index, body, doc_type='_doc', id=None, params=None, headers=None):
        if index not in self.__documents_dict:
            self.__documents_dict[index] = list()

        version = 1

        if id is None:
            id = get_random_id()
        elif self.exists(index, id, doc_type=doc_type, params=params):
            doc = self.get(index, id, doc_type=doc_type, params=params)
            version = doc['_version'] + 1
            self.delete(index, id, doc_type=doc_type)

        self.__documents_dict[index].append({
            '_type': doc_type,
            '_id': id,
            '_source': body,
            '_index': index,
            '_version': version
        })

        return {
            '_type': doc_type,
            '_id': id,
            'created': True,
            '_version': version,
            '_index': index
        }

    @query_params('consistency', 'op_type', 'parent', 'refresh', 'replication',
                  'routing', 'timeout', 'timestamp', 'ttl', 'version', 'version_type')
    def bulk(self, body, index=None, doc_type=None, params=None, headers=None):
        items = []
        errors = False

        for raw_line in body.splitlines():
            if len(raw_line.strip()) > 0:
                line = json.loads(raw_line)

                if any(action in line for action in ['index', 'create', 'update', 'delete']):
                    action = next(iter(line.keys()))

                    version = 1
                    index = line[action].get('_index') or index
                    doc_type = line[action].get('_type', "_doc")  # _type is deprecated in 7.x

                    if action in ['delete', 'update'] and not line[action].get("_id"):
                        raise RequestError(400, 'action_request_validation_exception', 'missing id')

                    document_id = line[action].get('_id', get_random_id())

                    if action == 'delete':
                        status, result, error = self._validate_action(
                            action, index, document_id, doc_type, params=params
                        )
                        item = {action: {
                            '_type': doc_type,
                            '_id': document_id,
                            '_index': index,
                            '_version': version,
                            'status': status,
                        }}
                        if error:
                            errors = True
                            item[action]["error"] = result
                        else:
                            self.delete(index, document_id, doc_type=doc_type, params=params)
                            item[action]["result"] = result
                        items.append(item)

                    if index not in self.__documents_dict:
                        self.__documents_dict[index] = list()
                else:
                    if 'doc' in line and action == 'update':
                        source = line['doc']
                    else:
                        source = line
                    status, result, error = self._validate_action(
                        action, index, document_id, doc_type, params=params
                    )
                    item = {
                        action: {
                            '_type': doc_type,
                            '_id': document_id,
                            '_index': index,
                            '_version': version,
                            'status': status,
                        }
                    }
                    if not error:
                        item[action]["result"] = result
                        if self.exists(index, document_id, doc_type=doc_type, params=params):
                            doc = self.get(index, document_id, doc_type=doc_type, params=params)
                            version = doc['_version'] + 1
                            self.delete(index, document_id, doc_type=doc_type, params=params)

                        self.__documents_dict[index].append({
                            '_type': doc_type,
                            '_id': document_id,
                            '_source': source,
                            '_index': index,
                            '_version': version
                        })
                    else:
                        errors = True
                        item[action]["error"] = result
                    items.append(item)
        return {
            'errors': errors,
            'items': items
        }

    def _validate_action(self, action, index, document_id, doc_type, params=None):
        if action in ['index', 'update'] and self.exists(index, id=document_id, doc_type=doc_type, params=params):
            return 200, 'updated', False
        if action == 'create' and self.exists(index, id=document_id, doc_type=doc_type, params=params):
            return 409, 'version_conflict_engine_exception', True
        elif action in ['index', 'create'] and not self.exists(index, id=document_id, doc_type=doc_type, params=params):
            return 201, 'created', False
        elif action == "delete" and self.exists(index, id=document_id, doc_type=doc_type, params=params):
            return 200, 'deleted', False
        elif action == 'update' and not self.exists(index, id=document_id, doc_type=doc_type, params=params):
            return 404, 'document_missing_exception', True
        elif action == 'delete' and not self.exists(index, id=document_id, doc_type=doc_type, params=params):
            return 404, 'not_found', True
        else:
            raise NotImplementedError(f"{action} behaviour hasn't been implemented")

    @query_params('parent', 'preference', 'realtime', 'refresh', 'routing')
    def exists(self, index, id, doc_type=None, params=None, headers=None):
        result = False
        if index in self.__documents_dict:
            for document in self.__documents_dict[index]:
                if document.get('_id') == id and document.get('_type') == doc_type:
                    result = True
                    break
        return result

    @query_params('_source', '_source_exclude', '_source_include', 'fields',
                  'parent', 'preference', 'realtime', 'refresh', 'routing', 'version',
                  'version_type')
    def get(self, index, id, doc_type='_all', params=None, headers=None):
        ignore = extract_ignore_as_iterable(params)
        result = None

        if index in self.__documents_dict:
            for document in self.__documents_dict[index]:
                if document.get('_id') == id:
                    if doc_type == '_all':
                        result = document
                        break
                    else:
                        if document.get('_type') == doc_type:
                            result = document
                            break

        if result:
            result['found'] = True
            return result
        elif params and 404 in ignore:
            return {'found': False}
        else:
            error_data = {
                '_index': index,
                '_type': doc_type,
                '_id': id,
                'found': False
            }
            raise NotFoundError(404, json.dumps(error_data))

    @query_params('_source', '_source_exclude', '_source_include',
                  'preference', 'realtime', 'refresh', 'routing',
                  'stored_fields')
    def mget(self, body, index, doc_type='_all', params=None, headers=None):
        ids = body.get('ids')
        results = []
        for id in ids:
            try:
                results.append(self.get(index, id, doc_type=doc_type,
                    params=params, headers=headers))
            except:
                pass
        if not results:
            raise RequestError(
                400,
                'action_request_validation_exception',
                'Validation Failed: 1: no documents to get;'
            )
        return {'docs': results}

    @query_params('_source', '_source_exclude', '_source_include', 'parent',
                  'preference', 'realtime', 'refresh', 'routing', 'version',
                  'version_type')
    def get_source(self, index, doc_type, id, params=None, headers=None):
        document = self.get(index=index, doc_type=doc_type, id=id, params=params)
        return document.get('_source')

    @query_params('_source', '_source_exclude', '_source_include',
                  'allow_no_indices', 'analyze_wildcard', 'analyzer', 'default_operator',
                  'df', 'expand_wildcards', 'explain', 'fielddata_fields', 'fields',
                  'from_', 'ignore_unavailable', 'lenient', 'lowercase_expanded_terms',
                  'preference', 'q', 'request_cache', 'routing', 'scroll', 'search_type',
                  'size', 'sort', 'stats', 'suggest_field', 'suggest_mode',
                  'suggest_size', 'suggest_text', 'terminate_after', 'timeout',
                  'track_scores', 'version')
    def count(self, index=None, doc_type=None, body=None, params=None, headers=None):
        searchable_indexes = self._normalize_index_to_list(index)

        i = 0
        for searchable_index in searchable_indexes:
            for document in self.__documents_dict[searchable_index]:
                if doc_type and document.get('_type') != doc_type:
                    continue
                i += 1
        result = {
            'count': i,
            '_shards': {
                'successful': 1,
                'skipped': 0,
                'failed': 0,
                'total': 1
            }
        }

        return result

    def _get_fake_query_condition(self, query_type_str, condition):
        return FakeQueryCondition(QueryType.get_query_type(query_type_str), condition)

    @query_params(
        "ccs_minimize_roundtrips",
        "max_concurrent_searches",
        "max_concurrent_shard_requests",
        "pre_filter_shard_size",
        "rest_total_hits_as_int",
        "search_type",
        "typed_keys",
    )
    def msearch(self, body, index=None, doc_type=None, params=None, headers=None):
        def grouped(iterable):
            if len(iterable) % 2 != 0:
                raise Exception('Malformed body')
            iterator = iter(iterable)
            while True:
                try:
                    yield (next(iterator)['index'], next(iterator))
                except StopIteration:
                    break

        responses = []
        took = 0
        for ind, query in grouped(body):
            response = self.search(index=ind, body=query)
            took += response['took']
            responses.append(response)
        result = {
            'took': took,
            'responses': responses
        }
        return result

    @query_params('_source', '_source_exclude', '_source_include',
                  'allow_no_indices', 'analyze_wildcard', 'analyzer', 'default_operator',
                  'df', 'expand_wildcards', 'explain', 'fielddata_fields', 'fields',
                  'from_', 'ignore_unavailable', 'lenient', 'lowercase_expanded_terms',
                  'preference', 'q', 'request_cache', 'routing', 'scroll', 'search_type',
                  'size', 'sort', 'stats', 'suggest_field', 'suggest_mode',
                  'suggest_size', 'suggest_text', 'terminate_after', 'timeout',
                  'track_scores', 'version')
    def search(self, index=None, doc_type=None, body=None, params=None, headers=None):
        searchable_indexes = self._normalize_index_to_list(index)

        matches = []
        conditions = []

        if body and 'query' in body:
            query = body['query']
            for query_type_str, condition in query.items():
                conditions.append(self._get_fake_query_condition(query_type_str, condition))
        for searchable_index in searchable_indexes:

            for document in self.__documents_dict[searchable_index]:

                if doc_type:
                    if isinstance(doc_type, list) and document.get('_type') not in doc_type:
                        continue
                    if isinstance(doc_type, str) and document.get('_type') != doc_type:
                        continue
                if conditions:
                    for condition in conditions:
                        if condition.evaluate(document):
                            matches.append(document)
                            break
                else:
                    matches.append(document)

        for match in matches:
            self._find_and_convert_data_types(match['_source'])

        if body and 'sort' in body:
            sort_query = body['sort']
            matches = self._sort_data(matches, sort_query)

        result = {
            'hits': {
                'total': {'value': len(matches), 'relation': 'eq'},
                'max_score': 1.0
            },
            '_shards': {
                # Simulate indexes with 1 shard each
                'successful': len(searchable_indexes),
                'skipped': 0,
                'failed': 0,
                'total': len(searchable_indexes)
            },
            'took': 1,
            'timed_out': False
        }

        hits = []
        for match in matches:
            match['_score'] = 1.0
            hits.append(match)

        # build aggregations
        if body is not None and 'aggs' in body:
            aggregations = {}

            for aggregation, definition in body['aggs'].items():
                aggregations[aggregation] = {
                    "doc_count_error_upper_bound": 0,
                    "sum_other_doc_count": 0,
                    "buckets": self.make_aggregation_buckets(definition, matches)
                }

            if aggregations:
                result['aggregations'] = aggregations

        if 'scroll' in params:
            result['_scroll_id'] = str(get_random_scroll_id())
            params['size'] = int(params.get('size', 10))
            params['from'] = int(params.get('from') + params.get('size') if 'from' in params else 0)
            self.__scrolls[result.get('_scroll_id')] = {
                'index': index,
                'doc_type': doc_type,
                'body': body,
                'params': params
            }
            hits = hits[params.get('from'):params.get('from') + params.get('size')]
        elif 'size' in params:
            hits = hits[:int(params['size'])]

        result['hits']['hits'] = hits

        return result

    @query_params('scroll')
    def scroll(self, scroll_id, params=None, headers=None):
        scroll = self.__scrolls.pop(scroll_id)
        result = self.search(
            index=scroll.get('index'),
            doc_type=scroll.get('doc_type'),
            body=scroll.get('body'),
            params=scroll.get('params')
        )
        return result

    @query_params('consistency', 'parent', 'refresh', 'replication', 'routing',
                  'timeout', 'version', 'version_type')
    def delete(self, index, id, doc_type=None, params=None, headers=None):

        found = False
        ignore = extract_ignore_as_iterable(params)

        if index in self.__documents_dict:
            for document in self.__documents_dict[index]:
                if document.get('_id') == id:
                    found = True
                    if doc_type and document.get('_type') != doc_type:
                        found = False
                    if found:
                        self.__documents_dict[index].remove(document)
                        break

        result_dict = {
            'found': found,
            '_index': index,
            '_type': doc_type,
            '_id': id,
            '_version': 1,
        }

        if found:
            return result_dict
        elif params and 404 in ignore:
            return {'found': False}
        else:
            raise NotFoundError(404, json.dumps(result_dict))

    @query_params('allow_no_indices', 'expand_wildcards', 'ignore_unavailable',
                  'preference', 'routing')
    def suggest(self, body, index=None, params=None, headers=None):
        if index is not None and index not in self.__documents_dict:
            raise NotFoundError(404, 'IndexMissingException[[{0}] missing]'.format(index))

        result_dict = {}
        for key, value in body.items():
            text = value.get('text')
            suggestion = int(text) + 1 if isinstance(text, int) else '{0}_suggestion'.format(text)
            result_dict[key] = [
                {
                    'text': text,
                    'length': 1,
                    'options': [
                        {
                            'text': suggestion,
                            'freq': 1,
                            'score': 1.0
                        }
                    ],
                    'offset': 0
                }
            ]
        return result_dict

    def _sort_data(self, matches, sort_query):
        if len(sort_query) == 0:
            return matches
        for sub_query in sort_query:
            if type(sub_query) is str:
                matches = sorted(matches, key=lambda item: item['_source'][sub_query])
            elif type(sub_query) is dict and len(sub_query) == 1:
                key, condition = list(sub_query.items())[0]
                if type(condition) is dict:
                    sort_type = condition.get('order', 'asc')
                else:
                    sort_type = condition
                if sort_type == 'desc':
                    matches = sorted(matches, key=lambda item: item['_source'][key], reverse=True)
                else:
                    matches = sorted(matches, key=lambda item: item['_source'][key])
            else:
                raise NotImplementedError(f"{sub_query} sorting has not been implemented")
        return matches

    def _normalize_index_to_list(self, index):
        # Ensure to have a list of index
        if index is None:
            searchable_indexes = self.__documents_dict.keys()
        elif isinstance(index, str) or isinstance(index, unicode):
            searchable_indexes = [index]
        elif isinstance(index, list):
            searchable_indexes = index
        else:
            # Is it the correct exception to use ?
            raise ValueError("Invalid param 'index'")

        # Check index(es) exists
        for searchable_index in searchable_indexes:
            if searchable_index not in self.__documents_dict:
                raise NotFoundError(404, 'IndexMissingException[[{0}] missing]'.format(searchable_index))

        return searchable_indexes

    @classmethod
    def _find_and_convert_data_types(cls, document):
        for key, value in document.items():
            if isinstance(value, dict):
                cls._find_and_convert_data_types(value)
            elif isinstance(value, datetime.datetime):
                document[key] = value.isoformat()

    def make_aggregation_buckets(self, aggregation, documents):
        if 'composite' in aggregation:
            return self.make_composite_aggregation_buckets(aggregation, documents)
        return []

    def make_composite_aggregation_buckets(self, aggregation, documents):

        def make_key(doc_source, agg_source):
            attr = list(agg_source.values())[0]["terms"]["field"]
            return doc_source[attr]

        def make_bucket(bucket_key, bucket):
            out = {
                "key": {k: v for k, v in zip(bucket_key_fields, bucket_key)},
                "doc_count": len(bucket),
            }

            for metric_key, metric_definition in aggregation["aggs"].items():
                metric_type_str = list(metric_definition)[0]
                metric_type = MetricType.get_metric_type(metric_type_str)
                attr = metric_definition[metric_type_str]["field"]
                data = [doc[attr] for doc in bucket]

                if metric_type == MetricType.CARDINALITY:
                    value = len(set(data))
                else:
                    raise NotImplementedError(f"Metric type '{metric_type}' not implemented")

                out[metric_key] = {"value": value}
            return out

        agg_sources = aggregation["composite"]["sources"]
        buckets = defaultdict(list)
        bucket_key_fields = [list(src)[0] for src in agg_sources]
        for document in documents:
            doc_src = document["_source"]
            key = tuple(make_key(doc_src, agg_src) for agg_src in aggregation["composite"]["sources"])
            buckets[key].append(doc_src)

        buckets = sorted(((k, v) for k, v in buckets.items()), key=lambda x: x[0])
        buckets = [make_bucket(bucket_key, bucket) for bucket_key, bucket in buckets]
        return buckets



# -*- coding: utf-8 -*-

from elasticmock import behaviour
from tests import TestElasticmock


class TestElasticmockBehaviour(TestElasticmock):

    def tearDown(self):
        behaviour.disable_all()



# -*- coding: utf-8 -*-

import setuptools

__version__ = '1.8.1'

# read the contents of your readme file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name='ElasticMockNew',
    version=__version__,
    author='t bittarn',
    author_email='t@bittarn.com',
    description='Python Elasticsearch Mock for test purposes with new features',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/t-bittarn/elasticmock',
    packages=setuptools.find_packages(exclude=('tests')),
    install_requires=[
        'elasticsearch',
        'mock',
        'python-dateutil',
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        "License :: OSI Approved :: MIT License",
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



# -*- coding: utf-8 -*-

from elasticsearch.exceptions import NotFoundError

from tests import TestElasticmock, INDEX_NAME, DOC_TYPE, BODY


class TestGet(TestElasticmock):

    def test_should_raise_notfounderror_when_nonindexed_id_is_used(self):
        with self.assertRaises(NotFoundError):
            self.es.get(index=INDEX_NAME, id='1')

    def test_should_not_raise_notfounderror_when_nonindexed_id_is_used_and_ignored(self):
        target_doc = self.es.get(index=INDEX_NAME, id='1', ignore=404)
        self.assertFalse(target_doc.get('found'))

    def test_should_not_raise_notfounderror_when_nonindexed_id_is_used_and_ignored_list(self):
        target_doc = self.es.get(index=INDEX_NAME, id='1', ignore=(401, 404))
        self.assertFalse(target_doc.get('found'))

    def test_should_get_document_with_id(self):
        data = self.es.index(index=INDEX_NAME, doc_type=DOC_TYPE, body=BODY)

        document_id = data.get('_id')
        target_doc = self.es.get(index=INDEX_NAME, id=document_id)

        expected = {
            '_type': DOC_TYPE,
            '_source': BODY,
            '_index': INDEX_NAME,
            '_version': 1,
            'found': True,
            '_id': document_id
        }

        self.assertDictEqual(expected, target_doc)

    def test_should_get_document_with_id_and_doc_type(self):
        data = self.es.index(index=INDEX_NAME, doc_type=DOC_TYPE, body=BODY)

        document_id = data.get('_id')
        target_doc = self.es.get(index=INDEX_NAME, id=document_id, doc_type=DOC_TYPE)

        expected = {
            '_type': DOC_TYPE,
            '_source': BODY,
            '_index': INDEX_NAME,
            '_version': 1,
            'found': True,
            '_id': document_id
        }

        self.assertDictEqual(expected, target_doc)

    def test_should_get_only_document_source_with_id(self):
        data = self.es.index(index=INDEX_NAME, doc_type=DOC_TYPE, body=BODY)

        document_id = data.get('_id')
        target_doc_source = self.es.get_source(index=INDEX_NAME, doc_type=DOC_TYPE, id=document_id)

        self.assertEqual(target_doc_source, BODY)

    def test_mget_get_several_documents_by_id(self):
        ids = []
        for _ in range(0, 10):
            data = self.es.index(index=INDEX_NAME, doc_type=DOC_TYPE, body=BODY)
            ids.append(data.get('_id'))
        results = self.es.mget(index=INDEX_NAME, body={'ids': ids})
        self.assertEqual(len(results['docs']), 10)
